/*
 * Copyright (C) 2002 Scott Smith (trckjunky@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <fcntl.h>
#include <errno.h>

#ifdef DIRECTISO
#include "player.h"
#else
#define genopen open
#define genclose close
#define genread read
#define genwrite write
#define genseek lseek64
#define tocgen_main main
#endif


// keeps TT_SRPT within 1 sector
#define MAXVTS 170

struct vtsdef {
	int numchapters,numsectors;
	char vtssummary[0x300],vtscat[4];
} vts[MAXVTS];
int numvts=0;

static unsigned int read2(unsigned char *p)
{
	return (p[0]<<8)|p[1];
}

static void write2(unsigned char *p,unsigned int foo)
{
	p[0]=foo>>8;
	p[1]=foo&255;
}

static unsigned int read4(unsigned char *p)
{
	return (p[0]<<24)|(p[1]<<16)|(p[2]<<8)|p[3];
}

static void write4(char *p,unsigned int v)
{
	p[0]=(v>>24)&255;
	p[1]=(v>>16)&255;
	p[2]=(v>>8)&255;
	p[3]=v&255;
}

static int toclen()
{
#ifdef DIRECTISO
	return 32;
#else
	return 2+(8+numvts*0x30c+2047)/2048;
#endif
}

static void TocGen(char *fname)
{
static char buf[2048];
int i,j;
int h;

	h=genopen(fname,O_WRONLY|O_CREAT|O_TRUNC,0644);

	memset(buf,0,2048);
	memcpy(buf,"DVDVIDEO-VMG",12);
	buf[0xf]=toclen()*2-1;
	buf[0x1f]=toclen()-1;
	buf[0x21]=0x11;
	buf[0x27]=1;
	buf[0x29]=1;
	buf[0x2a]=1;
	write2(buf+0x3e,numvts);
	memcpy(buf+0x40,"dvdauthor",9);
	buf[0x82]=5;
	buf[0x83]=0xb;
	buf[0x86]=4;
	buf[0xc7]=1;
	buf[0xd3]=2;
	buf[0x407]=0xc0;
	buf[0x4e5]=0xec;
	buf[0x4ed]=3;
	buf[0x4f3]=0x1f;
	buf[0x4f4]=0x71;
//	buf[0x4fc]=0x71; // r[0x01] = 0x0000
//	buf[0x4ff]=1;
	buf[0x4fc] = 0x41;
	buf[0x4ff] = 0x80;
	buf[0x504]=0x30;
	buf[0x505]=0x02;
	buf[0x509]=1;
	genwrite(h,buf,2048);


	// TT_SRPT
	// right now we assume 1 title per titleset, thus # titles == # titlesets
	memset(buf,0,2048);
	write2(buf,numvts); // # of titles
	write4(buf+4,7+12*numvts); // last byte of entry
	j=toclen()*2;
	for( i=0; i<numvts; i++ ) {
		buf[0x8+i*12]=0x3c; // title type
		buf[0x9+i*12]=0x1; // number of angles
		write2(buf+0xa+i*12,vts[i].numchapters); // number of chapters
		buf[0xe + i*12]=i+1; // VTS #
		buf[0xf+i*12]=1; // title # within VTS
		write4(buf+0x10+i*12,j); // start sector for VTS
		j+=vts[i].numsectors;
	}
	genwrite(h,buf,2048);

	// VMG_VTS_ATRT
	memset(buf,0,2048);
	j=8+numvts*4;
	write2(buf,numvts);
	write4(buf+4,numvts*0x30c+8-1);
	for( i=0; i<numvts; i++ )
		write4(buf+8+i*4,j+i*0x308);
	genwrite(h,buf,j);
	for( i=0; i<numvts; i++ ) {
		write4(buf,0x307);
		memcpy(buf+4,vts[i].vtscat,4);
		memcpy(buf+8,vts[i].vtssummary,0x300);
		genwrite(h,buf,0x308);
		j+=0x308;
	}
	j=2048-(j&2047);
	if( j < 2048 ) {
		memset(buf,0,j);
		genwrite(h,buf,j);
	}
	genclose(h);
}

static void ScanIfo(char *ifo)
{
	unsigned char buf[2048];
	int h=genopen(ifo,O_RDONLY);
	if( h<0 ) {
		fprintf(stderr,"cannot open %s: %s\n",ifo,strerror(errno));
		exit(1);
	}
	if( numvts+1 >= MAXVTS ) {
		fprintf(stderr,"Too many VTSs\n");
		exit(1);
	}
	memset(buf,0,2048);
	genread(h,buf,2048);
	vts[numvts].numsectors=read4(buf+0xc)+1;
	memcpy(vts[numvts].vtscat,buf+0x22,4);
	memcpy(vts[numvts].vtssummary,buf+0x100,0x300);
	memset(buf,0,2048);
	genread(h,buf,2048); // VTS_PTT_SRPT is 2nd sector
	vts[numvts].numchapters=read2(buf+read4(buf+4)-1);
	genclose(h);
	numvts++;
}

int tocgen_main(int argc,char **argv)
{
int i,nm=0;
static char fbuf[1000];
char temp[64];
int t;

	numvts=0;
	if( argc != 2 ) {
		fprintf(stderr,"Must pass in VIDEO_TS directory as a parameter\n");
		return 1;
	}
	fprintf(stderr,"tocgen running\n");

	for(i=0;i<100;++i)
	{
		sprintf(temp,"VTS_%02d_0.IFO",i);
		t=genopen(temp,O_RDONLY);
		if(t<0) continue;
		nm=i;
		genclose(t);
	}
	for( i=1; i<=nm; i++ ) {
		sprintf(fbuf,"%s/VTS_%02d_0.IFO",argv[1],i);
		fprintf(stderr,"scanning %s\n",fbuf);
		ScanIfo(fbuf);
	}
	if( !numvts ) {
		fprintf(stderr,"No .IFO files to process\n");
		return 1;
	}
	sprintf(fbuf,"%s/VIDEO_TS.IFO",argv[1]);
	TocGen(fbuf);
	sprintf(fbuf,"%s/VIDEO_TS.BUP",argv[1]);
	TocGen(fbuf);
	return 0;
}
